package code;

import graphicslib3D.*;
import graphicslib3D.GLSLUtils.*;
import graphicslib3D.shape.*; // Rolf

import java.nio.*;
import javax.swing.*;
import static com.jogamp.opengl.GL4.*;
import com.jogamp.opengl.*;
import com.jogamp.opengl.awt.GLCanvas;
import com.jogamp.opengl.util.*;
import com.jogamp.opengl.GLContext;
import com.jogamp.common.nio.Buffers;
import java.awt.event.KeyListener; // Rolf
import java.awt.event.KeyEvent; // Rolf
import java.util.*; // Rolf

public class Code extends JFrame implements GLEventListener,
                                            KeyListener // Rolf	
{	private GLCanvas myCanvas;
	private int rendering_program;
	private int vao[] = new int[1];
	private int vbo[] = new int[2];
	private float cameraX, cameraY, cameraZ, cameraXangle, cameraYangle;
	private float cubeLocX, cubeLocY, cubeLocZ;
	private GLSLUtils util = new GLSLUtils();
	private Sphere mySphere = new Sphere(24); // Rolf
	private	Matrix3D vMat = new Matrix3D(); // Rolf
        private Set<Integer> pressedKeys = new HashSet<>(10); // Rolf

	public Code()
	{	setTitle("Chapter4 - program1c");
		setSize(600, 600);
		myCanvas = new GLCanvas();
		myCanvas.addGLEventListener(this);
                myCanvas.addKeyListener(this); // Rolf
		getContentPane().add(myCanvas);
//		setFocusable(true); // Rolf (does not seem to give anything).
//              this.requestFocusInWindow();
		setVisible(true);
		FPSAnimator animator = new FPSAnimator(myCanvas, 30);
		animator.start();
	}

public synchronized void keyTyped(KeyEvent e) // Rolf
{}

public synchronized void keyPressed(KeyEvent e) // Rolf
{
	int keyCode = e.getKeyCode();
	pressedKeys.add(keyCode);
	Matrix3D newvMat = new Matrix3D();
        System.out.println("***");
	for (int keyCodeInSet: pressedKeys){
        System.out.println("Key in set	: " + keyCodeInSet);
        if (keyCodeInSet == KeyEvent.VK_RIGHT || keyCodeInSet == KeyEvent.VK_KP_RIGHT) {
	   newvMat.translate(-0.35f,0.0f,0.0f);
        }
        if (keyCodeInSet == KeyEvent.VK_LEFT || keyCodeInSet == KeyEvent.VK_KP_LEFT) {
	   newvMat.translate(0.35f,0.0f,0.0f);
        }
        if (keyCodeInSet == KeyEvent.VK_UP || keyCodeInSet == KeyEvent.VK_KP_UP) {
	   newvMat.translate(0.0f,0.0f,0.35f);
        }
        if (keyCodeInSet == KeyEvent.VK_DOWN || keyCodeInSet == KeyEvent.VK_KP_DOWN) {
	   newvMat.translate(-0.0f,0.0f,-0.35f);
        }
        if (keyCodeInSet == KeyEvent.VK_A) {
	   newvMat.rotateY(-0.5f);
        }
        if (keyCodeInSet == KeyEvent.VK_S) {
	   newvMat.rotateX(-0.5f);
        }
        if (keyCodeInSet == KeyEvent.VK_D) {
	   newvMat.rotateY(0.5f);
        }
        if (keyCodeInSet == KeyEvent.VK_W) {
	   newvMat.rotateX(0.5f);
        }
        if (keyCodeInSet == KeyEvent.VK_Z) {
	   newvMat.rotateZ(-0.5f);
        }
        if (keyCodeInSet == KeyEvent.VK_X) {
	   newvMat.rotateZ(0.5f);
        }
	}
	newvMat.concatenate(vMat);
        vMat = newvMat;
}

public synchronized void keyReleased(KeyEvent e){ // Rolf
	int keyCode = e.getKeyCode();
	pressedKeys.remove(keyCode);
}

	public void display(GLAutoDrawable drawable)
	{	GL4 gl = (GL4) GLContext.getCurrentGL();

		gl.glClear(GL_DEPTH_BUFFER_BIT);
		float bkg[] = { 0.0f, 0.0f, 0.0f, 1.0f };
		FloatBuffer bkgBuffer = Buffers.newDirectFloatBuffer(bkg);
		gl.glClearBufferfv(GL_COLOR, 0, bkgBuffer);

		gl.glUseProgram(rendering_program);

		int mv_loc = gl.glGetUniformLocation(rendering_program, "mv_matrix");
		int proj_loc = gl.glGetUniformLocation(rendering_program, "proj_matrix");

		float aspect = (float) myCanvas.getWidth() / (float) myCanvas.getHeight();
		Matrix3D pMat = perspective(60.0f, aspect, 0.1f, 1000.0f);

		// Rolf

		Matrix3D mMat = new Matrix3D();
		double x = (double) (System.currentTimeMillis())/10000.0;  // time factor
		mMat.translate(Math.sin(1*x)*2.0, Math.sin(1.5*x)*2.0, Math.sin(2*x)*2.0);
		mMat.rotate(500*x, 500*x, 500*x);

		Matrix3D mvMat = new Matrix3D();
		mvMat.concatenate(vMat);
		mvMat.concatenate(mMat);

		gl.glUniformMatrix4fv(mv_loc, 1, false, mvMat.getFloatValues(), 0);
		gl.glUniformMatrix4fv(proj_loc, 1, false, pMat.getFloatValues(), 0);

		gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
		gl.glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
		gl.glEnableVertexAttribArray(0);

		gl.glEnable(GL_DEPTH_TEST);
		gl.glDepthFunc(GL_LEQUAL);

		gl.glDrawArrays(GL_TRIANGLES, 0, 36);

		// Rolf begin - draw the sphere
		mMat = new Matrix3D();
		mMat.scale(24.0f, 12.0f, 12.0f);
		mvMat = new Matrix3D();
		mvMat.concatenate(vMat);
		mvMat.concatenate(mMat);

		gl.glUniformMatrix4fv(mv_loc, 1, false, mvMat.getFloatValues(), 0);
		gl.glUniformMatrix4fv(proj_loc, 1, false, pMat.getFloatValues(), 0);
		
		gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
		gl.glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
		gl.glEnableVertexAttribArray(0);

		int numVerts = mySphere.getIndices().length;
		gl.glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
		gl.glDrawArrays(GL_TRIANGLES, 0, numVerts);
		gl.glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
		// Rolf end

	}

	public void init(GLAutoDrawable drawable)
	{	GL4 gl = (GL4) GLContext.getCurrentGL();
		rendering_program = createShaderProgram();
		setupVertices();
		cameraX = 0.0f; cameraY = 0.0f; cameraZ = 8.0f;
		cubeLocX = 0.0f; cubeLocY = -2.0f; cubeLocZ = 0.0f;	// not used here
                cameraXangle = 0.0f; cameraYangle = 0.0f;
		vMat.translate(-cameraX,-cameraY,-cameraZ);
	}

	private void setupVertices()
	{	GL4 gl = (GL4) GLContext.getCurrentGL();
		float[] vertex_positions =
		{	-1.0f,  1.0f, -1.0f, -1.0f, -1.0f, -1.0f, 1.0f, -1.0f, -1.0f,
			1.0f, -1.0f, -1.0f, 1.0f,  1.0f, -1.0f, -1.0f,  1.0f, -1.0f,
			1.0f, -1.0f, -1.0f, 1.0f, -1.0f,  1.0f, 1.0f,  1.0f, -1.0f,
			1.0f, -1.0f,  1.0f, 1.0f,  1.0f,  1.0f, 1.0f,  1.0f, -1.0f,
			1.0f, -1.0f,  1.0f, -1.0f, -1.0f,  1.0f, 1.0f,  1.0f,  1.0f,
			-1.0f, -1.0f,  1.0f, -1.0f,  1.0f,  1.0f, 1.0f,  1.0f,  1.0f,
			-1.0f, -1.0f,  1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f,  1.0f,
			-1.0f, -1.0f, -1.0f, -1.0f,  1.0f, -1.0f, -1.0f,  1.0f,  1.0f,
			-1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  1.0f, -1.0f, -1.0f,
			1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f,
			-1.0f,  1.0f, -1.0f, 1.0f,  1.0f, -1.0f, 1.0f,  1.0f,  1.0f,
			1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f, -1.0f
		};

		Vertex3D[] vertices = mySphere.getVertices(); // Rolf
		int[] indices = mySphere.getIndices(); // Rolf
		
		float[] pvalues = new float[indices.length*3]; // Rolf
	
		for (int i=0; i<indices.length; i++) // Rolf
		{	pvalues[i*3] = (float) (vertices[indices[i]]).getX();
			pvalues[i*3+1] = (float) (vertices[indices[i]]).getY();
			pvalues[i*3+2] = (float) (vertices[indices[i]]).getZ();
		}
		
		gl.glGenVertexArrays(vao.length, vao, 0);
		gl.glBindVertexArray(vao[0]);
		gl.glGenBuffers(vbo.length, vbo, 0);

		gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
		FloatBuffer vertBuf = Buffers.newDirectFloatBuffer(vertex_positions);
		gl.glBufferData(GL_ARRAY_BUFFER, vertBuf.limit()*4, vertBuf, GL_STATIC_DRAW);

		// Rolf
		gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
		vertBuf = Buffers.newDirectFloatBuffer(pvalues);
		gl.glBufferData(GL_ARRAY_BUFFER, vertBuf.limit()*4, vertBuf, GL_STATIC_DRAW);
	}

	private Matrix3D perspective(float fovy, float aspect, float n, float f)
	{	float q = 1.0f / ((float) Math.tan(Math.toRadians(0.5f * fovy)));
		float A = q / aspect;
		float B = (n + f) / (n - f);
		float C = (2.0f * n * f) / (n - f);
		Matrix3D r = new Matrix3D();
		r.setElementAt(0,0,A);
		r.setElementAt(1,1,q);
		r.setElementAt(2,2,B);
		r.setElementAt(3,2,-1.0f);
		r.setElementAt(2,3,C);
		r.setElementAt(3,3,0.0f);
		return r;
	}

	public static void main(String[] args) { new Code(); }
	public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {}
	public void dispose(GLAutoDrawable drawable) {}

	private int createShaderProgram()
	{	GL4 gl = (GL4) GLContext.getCurrentGL();

		String vshaderSource[] = util.readShaderSource("code/vert.shader");
		String fshaderSource[] = util.readShaderSource("code/frag.shader");

		int vShader = gl.glCreateShader(GL_VERTEX_SHADER);
		int fShader = gl.glCreateShader(GL_FRAGMENT_SHADER);

		gl.glShaderSource(vShader, vshaderSource.length, vshaderSource, null, 0);
		gl.glShaderSource(fShader, fshaderSource.length, fshaderSource, null, 0);

		gl.glCompileShader(vShader);
		gl.glCompileShader(fShader);

		int vfprogram = gl.glCreateProgram();
		gl.glAttachShader(vfprogram, vShader);
		gl.glAttachShader(vfprogram, fShader);
		gl.glLinkProgram(vfprogram);
		return vfprogram;
	}
}
